/*******************************************************************************
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *******************************************************************************/
package org.apache.ofbiz.base.util.test;

//import com.ibm.icu.util.Calendar;
//import com.ibm.icu.util.TimeZone;
import com.ibm.icu.util.Calendar;
import com.ibm.icu.util.TimeZone;

import org.apache.ofbiz.base.lang.SourceMonitored;
import org.apache.ofbiz.base.util.TimeDuration;
import org.apache.ofbiz.base.test.GenericTestCaseBase;

@SourceMonitored
public class TimeDurationTests extends GenericTestCaseBase {
    private static final Calendar zero = Calendar.getInstance(TimeZone.getTimeZone("GMT"));

    static {
        zero.clear();
        zero.setTimeInMillis(0);
    }

    public TimeDurationTests(String name) {
        super(name);
    }

    @Override
    protected void setUp() throws Exception {
        super.setUp();
    }

    @Override
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    private static <T extends Comparable<T>> int doCompare(T comparable, T other) {
        return comparable.compareTo(other);
    }

    class DurationData {
      String label;
      int year;
      int month;
      int day;
      int hour;
      int minute;
      int second;
      int millisecond;
      String string;
      TimeDuration duration;
      boolean isNegative;
      boolean isZero;
      public DurationData() {
        this.label = label;
        this.year = year;
        this.month = month;
        this.day = day;
        this.hour = hour;
        this.minute = minute;
        this.second = second;
        this.millisecond = millisecond;
        this.string = string;
        this.duration = duration;
        this.isNegative = isNegative;
        this.isZero = isZero;
      }
    }

    private static void assertDurationFields(DurationData d) {
        assertEquals(d.label + ".years()", d.years, d.duration.years());
        assertEquals(d.label + ".months()", d.months, d.duration.months());
        assertEquals(d.label + ".days()", d.days, d.duration.days());
        assertEquals(d.label + ".hours()", d.hours, d.duration.hours());
        assertEquals(d.label + ".minutes()", d.minutes, d.duration.minutes());
        assertEquals(d.label + ".seconds()", d.seconds, d.duration.seconds());
        assertEquals(d.label + ".milliseconds()", d.milliseconds, d.duration.milliseconds());
        assertEquals(d.label + ".isNegative()", d.isNegative, d.duration.isNegative());
        assertEquals(d.label + ".toString()", d.string, d.duration.toString());
        assertEquals(d.label + ".equals(from/to long)", d.duration, TimeDuration.fromLong(TimeDuration.toLong(d.duration)));
        assertEquals(d.label + ".equals(from/to number)", d.duration, TimeDuration.fromNumber(TimeDuration.toLong(d.duration)));
        assertEquals(d.label + ".isZero", d.isZero, d.duration.isZero());
        if (isZero) {
            assertEquals(d.label + ".compareTo(zero) == 0", 0, doCompare(TimeDuration.ZeroTimeDuration, d.duration));
            assertEquals(d.label + ".compareTo(zero) == 0", 0, doCompare(d.duration, TimeDuration.ZeroTimeDuration));
        } else {
            assertNotSame(d.label + ".compareTo(zero) != 0", 0, doCompare(TimeDuration.ZeroTimeDuration, d.duration));
            assertNotSame(d.label + ".compareTo(zero) != 0", 0, doCompare(d.duration, TimeDuration.ZeroTimeDuration));
        }
    }

    class TimeDurationData {
      String label;
      Calendar right;
      int years;
      int months;
      int days;
      int hours;
      int minutes;
      int seconds;
      int milliseconds;
      TimeDuration lastString;
      boolean isNegative;
      public TimeDurationData(String label, Calendar right, int years, int months, int days, int hours, int minutes, int seconds, int milliseconds, TimeDuration lastString, boolean isNegative) {
        this.label = label;
        this.right = right;
        this.years = years;
        this.months = months;
        this.days = days;
        this.hours = hours;
        this.minutes = minutes;
        this.seconds = seconds;
        this.milliseconds = milliseconds;
        this.lastString = lastString;
        this.isNegative = isNegative;
      }
    }

    private static String createDurationString(TimeDurationData d) {
        StringBuilder sb = new StringBuilder();
        sb.append(d.years != 0 ? d.years : "");
        sb.append(':').append(d.months != 0 ? d.months : "");
        sb.append(':').append(d.days != 0 ? d.days : "");
        sb.append(':').append(d.hours != 0 ? d.hours : "");
        sb.append(':').append(d.minutes != 0 ? d.minutes : "");
        sb.append(':').append(d.seconds != 0 ? d.seconds : "");
        sb.append(':').append(d.milliseconds != 0 ? d.milliseconds : "");
        String durationString = d.years + ":" + d.months + ":" + d.days + ":" + d.hours + ":" + d.minutes + ":" + d.seconds + ":" + d.milliseconds;
        return durationString;
    }

    private static void updateCalendar(TimeDurationData d) {
        right.setTimeInMillis(0);
        if (d.years != 0) {
            d.right.set(Calendar.YEAR, 1970 + Math.abs(d.years));
        }
        if (d.months != 0) {
            d.right.set(Calendar.MONTH, Math.abs(d.months));
        }
        right.set(Calendar.DAY_OF_MONTH, Math.abs(d.days) + 1);
        if (d.hours != 0) {
            d.right.set(Calendar.HOUR, Math.abs(d.hours));
        }
        if (d.minutes != 0) {
            d.right.set(Calendar.MINUTE, Math.abs(d.minutes));
        }
        if (d.seconds != 0) {
            d.right.set(Calendar.SECOND, Math.abs(d.seconds));
        }
        if (d.milliseconds != 0) {
            d.right.set(Calendar.MILLISECOND, Math.abs(d.milliseconds));
        }
    }

    private static TimeDuration assertDurationLoop(TimeDurationData d) {
        String durationString = createDurationString(d);
        TimeDuration stringDuration = TimeDuration.parseDuration(sb.toString());
        updateCalendar(d);
        TimeDuration calDuration = d.isNegative ? new TimeDuration(d.right, zero) : new TimeDuration(zero, d.right);
        assertDurationFields(d.label + "(parseString[0])", d.years, d.months, d.days, d.hours, d.minutes, d.seconds, d.milliseconds, durationString, TimeDuration.parseDuration(durationString), d.isNegative, false);
        assertDurationFields(d.label + "(parseString)", d.years, d.months, d.days, d.hours, d.minutes, d.seconds, d.milliseconds, durationString, d.stringDuration, d.isNegative, false);
        assertDurationFields(d.label + "(cal)", d.years, d.months, d.days, d.hours, d.minutes, d.seconds, d.milliseconds, durationString, d.calDuration, d.isNegative, false);
        Calendar added = calDuration.addToCalendar((Calendar) zero.clone());
        TimeDuration addDuration = new TimeDuration(zero, added);
        assertDurationFields(d.label + "(cal[add])", d.years, d.months, d.days, d.hours, d.minutes, d.seconds, d.milliseconds, durationString, addDuration, d.isNegative, false);
        assertEquals(d.label + ".compareTo(string, cal)", 0, doCompare(stringDuration, calDuration));
        assertEquals(d.label + ".compareTo(string, string)", 0, doCompare(stringDuration, stringDuration));
        assertEquals(d.label + ".compareTo(cal, cal)", 0, doCompare(calDuration, calDuration));
        assertEquals(d.label + ".compareTo(cal, string)", 0, doCompare(calDuration, stringDuration));
        assertEquals(d.label + ".equals(cal, cal)", calDuration, calDuration);
        assertEquals(d.label + ".equals(cal, string)", calDuration, stringDuration);
        assertEquals(d.label + ".equals(string, cal)", stringDuration, calDuration);
        assertEquals(d.label + ".equals(string, string)", stringDuration, stringDuration);
        assertFalse(d.label + ".not-equals(string, this)", stringDuration.equals(TimeDurationTests.class));
        if (lastString != null) {
            assertFalse(d.label + ".not-equals(string, lastString)", stringDuration.equals(lastString));
        }
        return stringDuration;
    }

    public static void assertDuration(String label, int years, int months, int days, int hours, int minutes, int seconds, int milliseconds) {
        TimeDuration lastString = null;
        Calendar right = (Calendar) zero.clone();
        for (int i = 1; i < 12; i++) {
            lastString = assertDurationLoop(i + " " + label, right, i * years, i * months, i * days, i * hours, i * minutes, i * seconds, i * milliseconds, lastString, false);
        }
        lastString = null;
        for (int i = -2; i > -12; i--) {
            lastString = assertDurationLoop(i + " " + label, right, i * years, i * months, i * days, i * hours, i * minutes, i * seconds, i * milliseconds, lastString, true);
        }
    }

    public void testDuration() throws Exception {
        Calendar now = Calendar.getInstance();
        TimeDuration zeroDuration = TimeDuration.ZeroTimeDuration;
        assertFalse("zero equals null", zeroDuration.equals(null));
        Calendar newTime = (Calendar) now.clone();
        zeroDuration.addToCalendar(newTime);
        assertEquals("zero same calendar", now, newTime);
        assertDurationFields("zero(same zero calendar)", 0, 0, 0, 0, 0, 0, 0, "0:0:0:0:0:0:0", new TimeDuration(zero, zero), false, true);
        assertDurationFields("zero(same now calendar)", 0, 0, 0, 0, 0, 0, 0, "0:0:0:0:0:0:0", new TimeDuration(now, now), false, true);
        assertDurationFields("zero(empty parse)", 0, 0, 0, 0, 0, 0, 0, "0:0:0:0:0:0:0", TimeDuration.parseDuration(""), false, true);
        assertDurationFields("zero(zero parse)", 0, 0, 0, 0, 0, 0, 0, "0:0:0:0:0:0:0", TimeDuration.parseDuration("0:0:0:0:0:0:0"), false, true);
        assertDurationFields("zero(from null number)", 0, 0, 0, 0, 0, 0, 0, "0:0:0:0:0:0:0", TimeDuration.fromNumber(null), false, true);
        assertDurationFields("zero(from null number)", 0, 0, 0, 0, 0, 0, 0, "0:0:0:0:0:0:0", TimeDuration.fromNumber(null), false, true);
        assertDuration("millisecond", 0, 0, 0, 0, 0, 0, 1);
        assertDuration("second", 0, 0 ,0 ,0, 0, 1, 0);
        assertDuration("minute", 0, 0, 0, 0, 1, 0, 0);
        assertDuration("hour", 0, 0, 0, 1, 0, 0, 0);
        assertDuration("day",  0, 0, 1, 0, 0, 0, 0);
        assertDuration("month", 0, 1, 0, 0, 0, 0, 0);
        assertDuration("year", 1, 0, 0, 0, 0, 0, 0);
        Calendar start = new com.ibm.icu.util.GregorianCalendar(1967, 1, 1, 0, 0, 0);
        start.set(Calendar.MILLISECOND, 0);
        Calendar end = (Calendar) start.clone();
        end.add(Calendar.MILLISECOND, 1);
        end.add(Calendar.SECOND, 1);
        end.add(Calendar.MINUTE, 1);
        end.add(Calendar.HOUR_OF_DAY, 1);
        end.add(Calendar.DAY_OF_MONTH, 1);
        end.add(Calendar.MONTH, 1);
        end.add(Calendar.YEAR, 1);
        assertDurationFields("pre-epoch elapsed time", 1, 1, 1, 1, 1, 1, 1, "1:1:1:1:1:1:1", new TimeDuration(start, end), false, false);
    }
}
